#ifndef __PERF_H

#define __PERF_H

#include <iostream>
#include <time.h>
#include <sys/time.h>

using namespace std;

/// Encapsulates the timing data of a block
/** There is one static instance of this class for each profiled block. */
class perf
{
  static double stack[];
  static perf *list;
  static int level;

  perf *next;
  int registered, line;
  const char *function, *file;
  double ops, ///< number of operations performed
         bytes, ///< number of bytes (read + written)
         total, ///< cost : total time including time spent in children
         child, ///< time spent in children
         own,   ///< own time : total minus child, computed by perf::display
         own_percent ;

  void reset_()  { total=0; child=0; calls=0; ops=0; bytes=0;}
  
 public:
  int calls;
  static void display(double=0.95); ///< Displays a sorted list of the blocks where 95% of the time is spent (based on own time).
  static void reset();   ///< Clears all previously accumulated timings.
  
  perf() {}
  perf(int l, const char *fn, const char *fl):
    registered(0), next(NULL),
    line(l),function(fn),file(fl), 
    ops(0),bytes(0),total(0),child(0) {}
    
  int operator<(perf const &inf) const {return own>inf.own;}
  ostream& operator>>(ostream& out);
  typedef double *dbl_ptr;
  
  void enter(double nops, double nbytes);
  
  void leave(double time)
  {
    child += stack[level--];   // accumulate information left by children functions
    total += time;
    stack[level] += time;   // add up to time spent in parent function
  }
  
  /// A timer.
  /** An instance of this class is created when entering a profiled block.
      It remembers when it has been created. The destructor is implicitly called
      when leaving the block. It computes how much time has elapsed and updates
      the timing info for the block. */
  class timer
  {
    perf& info;
    timeval t;
    void gettime(timeval& tt) { gettimeofday(&tt, NULL); }

    //    timespec t;
    //    void gettime(timespec& tt) { clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &tt); }
    
    double elapsed()
    {
      //      timespec now; gettime(now);
      //      return double(now.tv_sec-t.tv_sec) + double(now.tv_nsec-t.tv_nsec)/1e9;
      timeval now; gettimeofday(&now, NULL);
      return double(now.tv_sec-t.tv_sec) + double(now.tv_usec-t.tv_usec)*1e-6;
    }
    static void init();
  public:
    timer(perf *inf, double ops, double bytes) : info(*inf)
      {
	gettime(t);
        info.enter(ops,bytes);
      }
    ~timer() { info.leave(elapsed()); }
    friend class perf;
  };
  
};

#define PERF(num,ops,bytes) \
  static perf info__(__LINE__, __PRETTY_FUNCTION__, __FILE__);	\
  perf::timer timer__(&info__,(num)*(ops),(num)*(bytes));

#endif
