#include <iostream>
#include <math.h>
#include <time.h>
#include <netcdfcpp.h>
using namespace std;

#include "perf.h"
#include "memory.h"
#include "params.h"

/**************************** Programme principal *****************************/

class Run
{
  enum { size=N*M };
  real x[M], y[M];
  array p0,u0,v0, p1,u1,v1, U,V,q,H;
  NcFile out;
  NcVar *pp,*qq,*t;

public :
  // Constructeur : filename = nom du fichier de sauvegarde
  Run(const char *filename);

  // Sauvegarde du champ de pression sur disque
  void save(array_in, real, int);

  // Condition initiale
  void init(array,array,array);

  // Execution d'un pas de temps : 
  // (p2,u2,v2) := (p0,u0,v0) + tau * F(p1,u1,v1)
  void step(real, array,array,array, array_in,array_in,array_in, array_in,array_in,array_in);

  // Boucle temporelle
  void go();    
};

int main()
{
  Run run("/tmp/saint_venant.nc");
  run.go();
  perf::display(2.0);
  return 0;
}

/********************************** Methodes **********************************/

array allocate(int size) { return new real[size]; }

Run::Run(const char *filename) : 
  U(allocate(size)), V(allocate(size)), H(allocate(size)), q(allocate(size)),
  p0(allocate(size)), p1(allocate(size)),  
  u0(allocate(size)), u1(allocate(size)),
  v0(allocate(size)), v1(allocate(size)), 
  out(filename, NcFile::Replace)
{
  // Creation et preparation du fichier de sauvegarde
  NcDim *xdim=out.add_dim("x",M), 
        *ydim=out.add_dim("y",N),
        *tdim=out.add_dim("t");
  NcVar *xx=out.add_var("x",ncFloat,xdim), 
        *yy=out.add_var("y",ncFloat,ydim);
  pp=out.add_var("p",ncFloat, tdim,xdim,ydim);
  qq=out.add_var("q",ncFloat, tdim,xdim,ydim);
  t=out.add_var("t",ncFloat,tdim);
  for(int i=0;i<M;i++) y[i]=dx*(i-M/2); yy->put(y,M);
  for(int i=0;i<N;i++) x[i]=dx*(i-N/2); xx->put(x,N);
}

void Run::save(array_in p, real tt, int k)
{
  PERF(M*N,0,sizeof(real));
  cout << "t=" << tt << '\n';
  t->set_cur(k); 
  t->put(&tt,1);
  pp->set_cur(k,0,0); 
  pp->put(p,1,M,N);
  qq->set_cur(k,0,0); 
  qq->put(q,1,M,N);
}

void Run::init(array p, array u, array v)
{
  for(int i=0,k=0; i<M; i++)
    for(int j=0; j<N; j++,k++)
      { 
	p[k] = c*c*(1+x[j]*exp(-.5*(x[j]*x[j]+y[i]*y[i]))); 
	u[k] = 0;
	v[k] = 0;
      }
}

void Run::step(real tau, array p2,array u2,array v2, 
	       array_in p1,array_in u1,array_in v1, array_in p0,array_in u0,array_in v0)
{
  PERF(0,0,0);
  dynp(H,p1,u1,v1);                          // H := p + .5*(u^2+v^2)
  pvort_massflux(q,U,V, p1,u1,v1);           // q := (f+dvdx-dudy)/p     U := p*u , V:=p*v
  periodize(U,U); periodize(V,V);
  stepp(tau,p2,p0, U,V);   periodize(p2,p2); // p2 = p2 - tau*(dUdx+dVdy)
  stepu(tau,u2,u0, q,V,H); periodize(u2,u2); // u2 = u0 - tau*(qV + dHdx)
  stepv(tau,v2,v0, q,U,H); periodize(v2,v2); // v2 = v0 + tau*(qU - dHdy)
}

void Run::go()
{
  PERF(0,0,0);
  init(p0,u0,v0);
  // premier pas de temps : (p1,phi1) := (p0,phi0) + dt * F(p0,phi0)
  step(dt, p1,u1,v1, p0,u0,v0, p0,u0,v0);

  real tt=0.;
  for (int k=0; k<P; k++)
    {
      save(p0,tt,k);
      for(int l=0; l<Q; l++)
	{
	  // pas de temps : (p0,phi0) := (p0,phi0) + 2*dt * F(p1,phi1)
	  step(2*dt, p0,u0,v0, p1,u1,v1, p0,u0,v0);
	  tt+=dt;
	  // pas de temps : (p1,phi1) := (p1,phi1) + 2*dt * F(p0,phi0)
	  step(2*dt, p1,u1,v1, p0,u0,v0, p1,u1,v1);
	  tt+=dt;
	}
    }
  save(p0,tt,P);
}
